// Modbus Server Mega board 
// Test for Simatic s7-1500 Modbus connections
// Vamos Sandor - 2020 - ob121.com
// https://www.ob121.com/doku.php?id=de:project_s7:s7_arduino_modbus_master

#include <Ethernet.h>
#include <ArduinoModbus.h>
#include <DHT.h>

#define DHTPIN 9 

// Uncomment whatever type you're using!
//#define DHTTYPE DHT11   // DHT 11
#define DHTTYPE DHT22   // DHT 22  (AM2302), AM2321
//#define DHTTYPE DHT21   // DHT 21 (AM2301)

DHT dht(DHTPIN, DHTTYPE);

word hum, temp, heatidx, lightint;
bool _2s;
unsigned long TimeAct, TimePrev, HoldingResult, InputResult, HeartBeat, i, StartingAddr;
long Cmd;

EthernetServer EthServer(502);
ModbusTCPServer modbusTCPServer;

void setup() {
  // Ethernet Settings
  byte mac[] = { 0x4E, 0xA0, 0xBE, 0x3F, 0xFE, 0x0F };  // Define MAc address
  byte ip[] = { 192, 168, 0, 42 };                      // Define IP address
  byte subnet[] = { 255, 255, 255, 0 };                 // Define SubNEt mask

  // initialize the ethernet device
  Ethernet.begin(mac, ip, subnet);                      // Assign MAC, IP, and subnet mask
  Serial.begin(9600);
  EthServer.begin();          // start listening for clients
  modbusTCPServer.begin();    // start listening for clients
  dht.begin();                // start dht
  
  // Define Holding register:
  HoldingResult = modbusTCPServer.configureHoldingRegisters(0, 100);
  InputResult = modbusTCPServer.configureInputRegisters(0, 100);

  Serial.print("Holding Reg init result =");
  Serial.print(HoldingResult);
  Serial.print("\n");

  Serial.print("Input Reg init result =");
  Serial.print(InputResult);
  Serial.print("\n");

  Serial.print("Modbus server address=");
  Serial.println(Ethernet.localIP());
  Serial.print("\n");
}

void loop() {
  // Modbus server accept incoming connections
  EthernetClient client = EthServer.available();

  if (client.connected()) {
    modbusTCPServer.accept(client);
    // poll for Modbus TCP requests, while client connected
    modbusTCPServer.poll();
    // Serial.print("poll");
  }

  //---------------------------------------------------------------------------------
  // TIME  clock 2s
  TimeAct = millis();     // Millis value between  0 and  655535

  _2s = false;
  if (  ( (TimeAct > TimePrev) and (TimeAct - TimePrev) > 2000) or ((TimeAct < TimePrev) and (65535 - TimePrev + TimeAct) > 2000 )  ) {
    _2s = true;
    TimePrev = TimeAct;
  }
  //---------------------------------------------------------------------------------
  // HeartBeat
  if (_2s) {
    HeartBeat++;
    if (HeartBeat > 255) {
      HeartBeat = 0;
    }
    Serial.print("HeartBeat=");
    Serial.println(HeartBeat);
    Serial.print("\n");
    dht22();                // call dht sensor  
    light();                // call photosensor
  }
  //---------------------------------------------------------------------------------
  // Modbus server  :

  // holding register 40001: heartbeat (FC3)
  modbusTCPServer.holdingRegisterWrite(0x00, HeartBeat);  // heart beat 2sec
  modbusTCPServer.holdingRegisterWrite(0x01, hum);        // humidity (/100, 0..100%, 2..5% accuracy)
  modbusTCPServer.holdingRegisterWrite(0x02, temp);       // temperature (/100,  -40..80 °C, 0,5% accuracy) 
  modbusTCPServer.holdingRegisterWrite(0x03, heatidx);    // heat index (/100, °C)
  modbusTCPServer.holdingRegisterWrite(0x04, lightint);   // light intensity (0..1024)
  modbusTCPServer.holdingRegisterWrite(0x05, 5);          // res
  modbusTCPServer.holdingRegisterWrite(0x06, 6);          // res

}

void dht22() {
  float h = dht.readHumidity();
  float t = dht.readTemperature();  
  float f = dht.readTemperature(true);
  // Check if any reads failed and exit early (to try again).
  if (isnan(h) || isnan(t) || isnan(f)) {
    Serial.println(F("Failed to read from DHT sensor!"));
    hum = 0;         // in errorcase: 0
    temp = 0;        // in errorcase: 0
    heatidx = 0;     // in errorcase: 0
    return;
  }

  // Compute heat index in Fahrenheit (the default)
  float hif = dht.computeHeatIndex(f, h);
  // Compute heat index in Celsius (isFahreheit = false)
  float hic = dht.computeHeatIndex(t, h, false);
 
  Serial.print(F("Humidity: "));
  Serial.print(h);
  Serial.print(F("%  Temperature: "));
  Serial.print(t);
  Serial.print(F("°C "));
  Serial.print(f);
  Serial.print(F("°F  Heat index: "));
  Serial.print(hic);
  Serial.print(F("°C "));
  Serial.print(hif);
  Serial.println(F("°F"));  

  h = h * 100.0;
  t = t * 100.0;
  hic = hic * 100.0;

  hum = (word) h;
  temp = (word) t;
  heatidx = (word) hic; 
}

void light() {
  // light sensor (photosensor): GL5506 (light resisitance (10 lux): 2-6 KOhm,
  // dark resistance 0.15 MOhm
  
  int val = analogRead(5);             // read from A5
  Serial.print("Light intensity : " );
  Serial.println(val); 
  lightint = (word) val;
}
